// Copyright 2009 Xerox Corporation.  All rights reserved.
// Apache 2.0 License
// See file copying.txt

package com.xerox.adoc.locatingrules;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URISyntaxException;
import java.util.HashMap;
import org.xml.sax.SAXException;

import java.util.Map;
import java.io.File;
import org.xml.sax.helpers.XMLFilterImpl;
import org.xml.sax.Attributes;

// This class assumes that the locating rules file ("schemas.xml") applies only to the current directory.
// This class implements only transformURI and partially implements uri elements.
// It does not implement schemas.xml document order priority rules.
// <uri resource="config.xml" uri="../../tools/rnc/config.rnc"/> does not require directory scanning
// but all other uses do.
// A better implementation would build a list of rules, and note whether any of the rules require directory scanning,
// and use directory scanning only if necessary.

class LocatingRulesParser extends XMLFilterImpl {
  private static final String LOCATING_RULES_NAMESPACE = "http://thaiopensource.com/ns/locating-rules/1.0";
  private static boolean DEBUG = false;

  private final Map<File, File> srcTargetFiles; // src->target
  private final File dir;

  LocatingRulesParser(File dir) {
    super();
    this.dir = dir;
    srcTargetFiles = new HashMap<File, File>();
  }

  public void startDocument() throws SAXException {
    super.startDocument();
  }

  public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
    if (!uri.equals(LOCATING_RULES_NAMESPACE))
      throw new SAXException("Did not expect namespace " + uri);
    if (localName.equals("locatingRules")) {
      // ok
    } else if (localName.equals("include")) {
      throw new SAXException(localName + " element not supported");
    } else if (localName.equals("applyFollowingRules")) {
      throw new SAXException(localName + " element not supported");
    } else if (localName.equals("documentElement")) {
      throw new SAXException(localName + " element not supported");
    } else if (localName.equals("namespace")) {
      throw new SAXException(localName + " element not supported");
    } else if (localName.equals("uri")) {
      handleURIElement(uri, localName, qName, atts);
    } else if (localName.equals("transformURI")) {
      handleTransformURI(uri, localName, qName, atts);
    } else if (localName.equals("default")) {
      throw new SAXException(localName + " element not supported");
    } else if (localName.equals("typeId")) {
      throw new SAXException(localName + " element not supported");
    } else if (localName.equals("expansionRule")) {
      throw new SAXException(localName + " element not supported");
    } else {
      throw new SAXException(localName + " element not supported");
    }
    super.startElement(uri, localName, qName, atts);
  }

  Map<File, File> getTargetFiles() {
    return srcTargetFiles;
  }

  private File resolveFileURI(File dir, String rnc) throws SAXException {
    if (rnc !=null && rnc.startsWith("file://")) {
      URL url;
      try {
        url = new URL(rnc);
      } catch (MalformedURLException e) {
        throw new SAXException("unable to parse uri=" + rnc);
      }
      try {
        return new File(url.toURI());
      } catch (URISyntaxException e) {
        return new File(url.getPath());
      }
    } else {
      if (DEBUG) System.out.println("dir=" + dir+" rnc="+rnc);
      return new File(dir, rnc);
    }
  }

//  transformURI =
//  ## Generates a URI for the related resource by transforming
//  ## the URI of the document. Matches if the transformation
//  ## yields a valid URI that identifies an existing resource.
//  element transformURI {
//    common,
//    ## Semantics are the same as the pattern attribute of the uri element.
//    attribute fromPattern { uriPattern },
//    ## The result of the transformation is produced from the toPattern
//    ## by replacing each * by the string that matched the corresponding
//    ## * in the toPattern.  The toPattern is appended to the initial
//    ## part of the document's URI that was not explicitly matched
//    ## by fromPattern.
//    attribute toPattern { uriPattern }
//  }

  private String requiredAttribute(String localName, Attributes atts, String attributeName) throws SAXException {
    String value = atts.getValue(attributeName);
    if (value == null) throw new SAXException(localName + " attribute " + attributeName + " required");
    return value;
  }

  private void unsupportedAttribute(String localName, Attributes atts, String attributeName) throws SAXException {
    String value = atts.getValue(attributeName);
    if (value != null) throw new SAXException(localName + " attribute " + attributeName + " not supported");
  }

  public void handleTransformURI(String uri, String localName, String qName, Attributes atts) throws SAXException {
    // <transformURI fromPattern="*.xml" toPattern="*.rnc" />
    String fromPattern = requiredAttribute(localName, atts, "fromPattern");
    String toPattern = requiredAttribute(localName, atts, "toPattern");

    for (File file : dir.listFiles(new Glob.FileNamePatternFilter(fromPattern))) {
      File toFile = Glob.replaceFilePattern(file, fromPattern, toPattern);
      if (DEBUG) System.out.println("file="+file+" toFile="+toFile+" fromPattern="+fromPattern+" toPattern="+toPattern);
      if (file.exists() && ! srcTargetFiles.containsKey(file))
        srcTargetFiles.put(file, toFile);
    }
  }


  public void handleURIElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
    unsupportedAttribute(localName, atts, "typeId");

    String rnc = requiredAttribute(localName, atts, "uri");
    File rncFile = resolveFileURI(dir, rnc);

    // <uri resource="config.xml" uri="../../tools/rnc/config.rnc"/>
    {
      String src = atts.getValue("resource");
      if (src != null) {
        File file = resolveFileURI(dir, src);
        if (file.exists() && ! srcTargetFiles.containsKey(file))
          srcTargetFiles.put(file, rncFile);
        return;
      }
    } 

    // <uri pattern="*.xml" uri="../../tools/rnc/config.rnc"/>
    {
      String pattern = atts.getValue("pattern");
      if (pattern != null) {
        for (File file : dir.listFiles(new Glob.FileNamePatternFilter(pattern))) {
          if (file.exists() && ! srcTargetFiles.containsKey(file))
            srcTargetFiles.put(file, rncFile);
        }
        return;
      }
    }
    throw new SAXException(localName + " element without resource or pattern");
  }

}
