// Copyright 2009 Xerox Corporation.  All rights reserved.
// Apache 2.0 License
// See file copying.txt

package com.xerox.adoc.locatingrules;

// adapted from com.thaiopensource.relaxng.util.JingTask
// whose copyright notice is reproduced below:
// Jing Copying Conditions
// 
// Copyright (c) 2001-2003 Thai Open Source Software Center Ltd
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
// * Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// * Neither the name of the Thai Open Source Software Center Ltd nor the
// names of its contributors may be used to endorse or promote products
// derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
// USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 

import com.thaiopensource.util.PropertyMapBuilder;
import com.thaiopensource.validate.Flag;
import com.thaiopensource.validate.SchemaReader;
import com.thaiopensource.validate.prop.rng.RngProperty;
import com.thaiopensource.validate.prop.schematron.SchematronProperty;
import com.thaiopensource.validate.rng.CompactSchemaReader;
import com.thaiopensource.xml.sax.ErrorHandlerImpl;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;

/**
 * Ant task to validate XML files using RELAX NG or other schema languages
 * under direction of one schemas.xml file per directory.  
 */

public class LocatingRulesTask extends Task {
  static final boolean DEBUG = false;
  
  private File schemasFile;
  private final Collection<FileSet> filesets = new ArrayList<FileSet>();
  private PropertyMapBuilder properties = new PropertyMapBuilder();
  private boolean failOnError = true;
  private boolean verbose = false;
  private boolean compactsyntax = false;
  private SourceDirectories sourceDirectories;

  private class LogErrorHandler extends ErrorHandlerImpl {
    int logLevel = Project.MSG_ERR;

    public void warning(SAXParseException e) throws SAXParseException {
      logLevel = Project.MSG_WARN;
      super.warning(e);
    }

    public void error(SAXParseException e) {
      logLevel = Project.MSG_ERR;
      super.error(e);
    }

    public void printException(Throwable e) {
      logLevel = Project.MSG_ERR;
      super.printException(e);
    }

    public void print(String message) {
      log(message, logLevel);
    }
  }

  public LocatingRulesTask() {
  }

  public void execute() throws BuildException {
    ErrorHandlerImpl eh = new LogErrorHandler();
    if ((schemasFile == null) == (filesets.isEmpty()))
      throw new BuildException("There must be either a file attribute or a fileset child element, but not both",
                               getLocation());

    SchemaReader schemaReader = (compactsyntax ? CompactSchemaReader.getInstance() : null);
    sourceDirectories = new SourceDirectories(properties, schemaReader);

    boolean hadError = false;
    try {
      // load locating rules
      if (schemasFile != null) {
        sourceDirectories.loadSchemas(schemasFile.getParentFile(), schemasFile);
      } else {
        loadSchemas(filesets);
      }
      // Validate
      hadError |= sourceDirectories.validate(eh, verbose);
    } catch (SAXException e) {
      hadError = true;
      eh.printException(e);
    } catch (IOException e) {
      hadError = true;
      eh.printException(e);
    }
    if (hadError && failOnError)
      throw new BuildException("Validation failed; messages should have been provided.", getLocation());
  }

  private void loadSchemas(Collection<FileSet> filesets) throws SAXException, IOException {
    for (FileSet fs : filesets) {      
      DirectoryScanner ds = fs.getDirectoryScanner(getProject());
      File dir = fs.getDir(getProject());
      for (String locatingRulesFile : ds.getIncludedFiles()) {
        sourceDirectories.loadSchemas(dir, new File(dir, locatingRulesFile));
      }
    }
  }

  /**
   * Handles the <code>checkid</code> attribute.
   *
   * @param checkid the attribute value converted to a boolean
   */
  public void setCheckid(boolean checkid) {
    properties.put(RngProperty.CHECK_ID_IDREF,
                   checkid ? Flag.PRESENT : null);
  }

  /**
   * Handles the <code>compactsyntax</code> attribute.
   *
   * @param compactsyntax the attribute value converted to a boolean
   */
  public void setCompactsyntax(boolean compactsyntax) {
    this.compactsyntax = compactsyntax;
  }

  /**
   * Handles the <code>feasible</code> attribute.
   *
   * @param feasible the attribute value converted to a boolean
   */
  public void setFeasible(boolean feasible) {
    properties.put(RngProperty.FEASIBLE, feasible ? Flag.PRESENT : null);
  }

  /**
   * Handles the phase attribute.
   *
   * @param phase the attribute value
   */
  public void setPhase(String phase) {
    SchematronProperty.PHASE.put(properties, phase);
  }

  /**
   * Handles the <code>failonerror</code> attribute.
   *
   * @param failOnError the attribute value converted to a boolean
   */
  public void setFailonerror(boolean failOnError) {
    this.failOnError = failOnError;
  }


  /**
   * Handles the ant <code>fileset</code>.
   *
   * @param set an ant FileSet of locatingRules files.
   */
  public void addFileset(FileSet set) {
    filesets.add(set);
  }

  /**
   * Handles the <code>schemas</code> attribute.
   *
   * @param file a single locatingRules file to be used instead of a fileset.
   */
  public void setSchemas(File file) {
    this.schemasFile = file;
  }

  /**
   * Handles the <code>verbose</code> attribute.
   *
   * @param verbose If true, prints out information about each file validated.  If false, prints only errors.
   */
  public void setVerbose(boolean verbose) {
    this.verbose = verbose;
  }

}
